import nltk
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


class StringToTreeTransformer(BaseEstimator, TransformerMixin):
    """
    Parses tree representations into their nltk.Tree object form.

    input: list of list of strings
    output: list of list of nltk.Tree objects
    """

    def fit(self, x, y=None):
        return self

    def transform(self, x, y=None):
        result = []
        for document in x:
            ret = []
            if isinstance(document, str):
                raise Exception(
                    'this transformer only takes document which '
                    'are already split into sentences. Please use a sentence '
                    'splitter before (e.g., by using stanfordnlp)'
                )
            for line in document:
                if line and line.strip():
                    tree = nltk.Tree.fromstring(line)
                    ret.append(tree)
            result.append(ret)
        return result


class WordToPosTransformer(BaseEstimator, TransformerMixin):
    """
    Creates a POS-only representation of a tree sentence.

    input: list of list of trees
    output: list of list of strings
    """

    def fit(self, x, y=None):
        return self

    def transform(self, x, y=None):
        ret = []
        for document in x:
            tmp = []
            for tree in document:
                tmp += [x[1] for x in tree.pos()]
            ret.append(tmp)
        return ret


class TreeChainTransformer(BaseEstimator, TransformerMixin):
    """
    Creates chains of strings with an optional length limit from nltk tree
    structures.

    For each leaf node, its path to the root node is called a chain.
    This transformer returns every subchain with a certain length from every
    chain. Common ancester subchains are only returned once.
    For example, in the tree (1 (2 (3, 4))) and max_lenth=2, the subchain 1-2
    is only returned once, although it is part of both chains 1-2-3 and 1-2-4.

    input document: list of nltk.Tree objects (one for each sentence)

    output document: the format of the output documents depend on the
    combination parameters.

    Args:
        max_length (int): the maximum length of the produced chains. If set to
            None, only full chains of any length will be returned.

        combine_chain_elements (str): if set to any string other than None,
            each chain will be returned as a string (e.g., \"PRON VERB NOUN\")
            instead of a list. The parameter is used to join the chains.
            Defaults to None.

        combine_chains (str): if set to true, return each document as the
            concatenation of all chains for all trees in the document. If set
            to None, returns a list of sentences for each document, where each
            sentence conists of a list of chains. The parameter is used to join
            the chains. Defaults to None.gv

        combine_strings (str): if set, it will be used to join the tree chains
            together to a single string per document. The parameter is used to
            join the chains. Defaults to None.


    """

    def __init__(
        self,
        max_length=None,
        combine_chain_elements=None,
        combine_chains=None,
        combine_strings=None,
    ):

        if combine_strings is not None and combine_chains is None:
            raise ValueError(
                'if combine_strings is set, both combine_chains and '
                'combine_chain_elementsmust be set too.'
            )
        if combine_chains is not None and combine_chain_elements is None:
            raise ValueError(
                'if combine_chains is set, combine_chain_elements must be set '
                'too.'
            )

        self.max_length = max_length
        self.combine_chain_elements = combine_chain_elements
        self.combine_chains = combine_chains
        self.combine_strings = combine_strings

    def fit(self, x, y=None):
        return self

    # please be careful, it is dangerous in here!
    def _find_chains(self, tree, prev_labels):
        ret_full = []
        if isinstance(tree, str) or len(tree) == 0:
            # the leaf nodes of nltk trees are strings
            if not self.max_length or len(prev_labels) < self.max_length:
                ret_full.append(prev_labels[:])
            return ret_full
        if self.max_length and len(prev_labels) == self.max_length:
            # case: intermediary subchain that stops before the root
            prev_labels = prev_labels[1:]

        prev_labels.append(tree.label())
        if len(prev_labels) == self.max_length:
            # by adding this node, we have completed a subchain
            ret_full.append(prev_labels[:])
        for child in tree:
            ret_full += self._find_chains(child, prev_labels[:])
        return ret_full

    def transform(self, x, y=None):
        result = []
        for document in x:
            #  a document consists of a list of nltk.Tree objects
            #  one tree represents one sentence.
            new_document = []
            for tree in document:
                #  every tree is split into a list of chains, whereas
                #  every chain is a list of symbols.
                chains = self._find_chains(tree, [])

                if self.combine_chain_elements is not None:
                    chains = [self.combine_chain_elements.join(x) for x in
                              chains]

                if self.combine_chains is not None:
                    new_document.append(self.combine_chains.join(chains))
                else:
                    new_document.append(chains)

            if self.combine_strings is not None:
                new_document = self.combine_strings.join(new_document)
            result.append(new_document)
        return result


class TreeStatsVectorizer(TransformerMixin, BaseEstimator):
    """
    calculates high-level, low-complexity tree features

    input: list of list of trees
    output: list of list of numbers/floats
    """

    def fit(self, x, y=None):
        return self

    def transform(self, x, y=None):
        result = []
        for document in x:
            result.append(
                [
                    self.get_average_height(document),
                    self.get_average_children(document),
                    self.get_average_inner_to_leaf_ratio(document),
                    self.get_max_child_width(document),
                ]
            )
        return result

    def get_average_height(self, document):
        return np.average([tree.height() for tree in document])

    def _calculate_average_children(self, tree):
        result = []
        if type(tree) != nltk.tree.Tree or not tree:
            result.append(0)
        else:
            result.append(len(tree))
            for child in tree:
                result += self._calculate_average_children(child)
        return result

    def get_average_children(self, document):
        result = []
        for tree in document:
            result += self._calculate_average_children(tree)
        return np.average(result)

    def get_average_inner_to_leaf_ratio(self, document):
        result = []
        for tree in document:
            num_leaves = len(tree.leaves())
            num_total = len(tree.treepositions()) - 1  # root node
            result.append(num_leaves / num_total)
        return np.average(result)

    def _get_max_tree_width(self, tree):
        maximum = 0
        if type(tree) == nltk.tree.Tree:
            for child in tree:
                maximum = max(maximum, self._get_max_tree_width(child),
                              len(tree))
        return maximum

    def get_max_child_width(self, document):
        result = []
        for tree in document:
            result.append(self._get_max_tree_width(tree))
        return np.average(result)


class TreeToStringTransformer(BaseEstimator, TransformerMixin):
    """
    Transforms nltk trees into strings that can be parsed later

    input: list of list of trees
    output: list of list of strings
    """

    def fit(self, x, y=None):
        return self

    def transform(self, x, y=None):
        ret = []
        for document in x:
            doc = []
            for tree in document:
                doc.append(tree.pformat(margin=1000000))
            ret.append(doc)
        return ret


class TreeToFlatStringTransformer(TransformerMixin, BaseEstimator):
    """
    Transforms NLTK trees into Strings by traversing the tree post-order.

    Input: Every document is a list of NLTK trees, one representing each
        sentence.
    Output: Every document is a single string.
    """

    def fit(self, X, y=None, **fit_args):
        return self

    def parse(self, document):
        result = [document.label()]
        result += [self.parse(child) for child in document]
        return ' '.join(result)

    def transform(self, X):
        result = []
        for document in X:
            parses = [self.parse(sentence) for sentence in document]
            result.append('. '.join(parses))
        return result



from collections import defaultdict

import numpy as np
import logging
logger = logging.getLogger(__name__)

from sklearn.base import BaseEstimator, TransformerMixin



def presence_kernel(x, y, ngram_min, ngram_max):
    result = np.zeros((len(x), len(y)), dtype=int)
    for i, string in enumerate(x):
        if type(string) == str:
            string = string.lower()
        ngrams1 = set()
        for ngram_size in range(ngram_min, ngram_max + 1):
            for index in range(len(string) - ngram_size + 1):
                ngram = string[index: index + ngram_size]
                if type(ngram) != str:
                    ngram = str(ngram)
                ngrams1.add(ngram)
        for j, counterstring in enumerate(y):
            if type(counterstring) == str:
                counterstring = counterstring.lower()
            ngrams2 = set()
            for ngram_size in range(ngram_min, ngram_max + 1):
                for index in range(len(counterstring) - ngram_size + 1):
                    ngram = counterstring[index: index + ngram_size]
                    if type(ngram) != str:
                        ngram = str(ngram)
                    ngrams2.add(ngram)
            result[i, j] = len(ngrams1.intersection(ngrams2))
    return result


def spectrum_kernel(x, y, ngram_min, ngram_max):
    result = np.zeros((len(x), len(y)), dtype=int)
    for i, string in enumerate(x):
        if type(string) == str:
            string = string.lower()
        ngrams = defaultdict(int)
        for ngram_size in range(ngram_min, ngram_max + 1):
            for index in range(len(string) - ngram_size + 1):
                ngram = string[index: index + ngram_size]
                if type(ngram) != str:
                    ngram = str(ngram)
                ngrams[ngram] += 1
        for j, counterstring in enumerate(y):
            if type(counterstring) == str:
                counterstring = counterstring.lower()
            for ngram_size in range(ngram_min, ngram_max + 1):
                for index in range(len(counterstring) - ngram_size + 1):
                    ngram = counterstring[index: index + ngram_size]
                    if type(ngram) != str:
                        ngram = str(ngram)
                    result[i, j] += ngrams[ngram]
    return result


def intersection_kernel(x, y, ngram_min, ngram_max):
    result = np.zeros((len(x), len(y)), dtype=int)
    for i, string in enumerate(x):
        if type(string) == str:
            string = string.lower()
        ngrams = defaultdict(int)
        for ngram_size in range(ngram_min, ngram_max + 1):
            for index in range(len(string) - ngram_size + 1):
                ngram = string[index: index + ngram_size]
                if type(ngram) != str:
                    ngram = str(ngram)
                ngrams[ngram] += 1
        for j, counterstring in enumerate(y):
            if type(counterstring) == str:
                counterstring = counterstring.lower()
            ngrams2 = dict(ngrams)
            for ngram_size in range(ngram_min, ngram_max + 1):
                for index in range(len(counterstring) - ngram_size + 1):
                    ngram = counterstring[index: index + ngram_size]
                    if type(ngram) != str:
                        ngram = str(ngram)
                    if ngram in ngrams2 and ngrams2[ngram] > 0:
                        result[i, j] += 1
                        ngrams2[ngram] -= 1
    return result


kernel_map = {
    "presence": presence_kernel,
    "spectrum": spectrum_kernel,
    "intersection": intersection_kernel,
}


class StringKernelTransformer(BaseEstimator, TransformerMixin):
    """
    Converts (string) documents to a similarity matrix (kernel).

    Input (fit): List of m strings
    Input (transform): List of n strings
    Output: m x n matrix containing the kernel similarities between the strings
    """

    def __init__(self, kernel_type="intersection", ngram_range=None,
                 normalize=True):
        if ngram_range is None:
            ngram_range = (5, 10)
        self.ngram_range = ngram_range
        self.normalize = normalize
        self.kernel_type = kernel_type
        if kernel_type not in kernel_map:
            raise ValueError(f"unknown kernel: {kernel_type}")
        self._kernel = kernel_map[kernel_type]
        self._train_data = None
        self._train_kernel = None

    def fit(self, X, y=None):
        self._train_data = np.array(X)
        self._train_kernel = self._kernel(X, X, *self.ngram_range)
        for i in range(len(X)):
            if self._train_kernel[i, i] == 0:
                logger.error(f"zeros in diagonal at ({i},{i}) for {X[i]}")
        return self

    def transform(self, X, y=None):
        X = np.array(X)
        _st = self._kernel(X, self._train_data, *self.ngram_range)

        if not self.normalize:
            return _st

        _ss = self._train_kernel
        _tt = self._kernel(X, X, *self.ngram_range)

        result = np.copy(_st)
        for i in range(_st.shape[0]):
            if _tt[i, i] == 0:
                logger.error(f"zeros in diagonal at ({i},{i}) for {X[i]}")
            for j in range(_st.shape[1]):
                if _tt[i, i] != 0 and _ss[j, j] != 0:
                    result[i, j] /= np.sqrt(_tt[i, i] * _ss[j, j])

        return result



"""Transformers that use stanza documents."""

import json
import logging
import os
import warnings

import nltk
import stanza
from sklearn.base import BaseEstimator, TransformerMixin
from stanza import Document
from stanza.models.common.doc import Sentence
from tqdm import tqdm

logger = logging.getLogger(__name__)

UNKNOWN_KEY = 'UNK'
WORD_FEATURES_JSON = os.path.join('resources', 'word_features.json')


def _get_word_attribute(word, attribute, fallback='_'):
    if fallback is None:
        raise ValueError("Fallback can't be None")
    if hasattr(word, attribute):
        a = getattr(word, attribute)
        if a is not None:
            return a
    if not hasattr(word, '_feats'):
        return fallback

    feats = word.feats
    if not feats or feats == '_':
        return fallback
    pairs = feats.split('|')
    features = {k: v for (k, v) in [x.split('=') for x in pairs]}
    if attribute not in features or features[attribute] is None:
        return fallback
    return features[attribute]


def _get_features_per_word(document):
    result = []
    words = []
    for sentence in document.sentences:
        words += sentence.words
    for word in words:
        features = word.feats
        if not features or features == '_':
            continue
        pairs = features.split('|')
        result.append({k: v for (k, v) in [x.split('=') for x in pairs]})
    return result


class StanzaWordFeatureFrequencyTransformer(BaseEstimator, TransformerMixin):
    """
    Returns a frequency matrix for all grammar features for each word of each
    document.

    This transformer looks for the file resources/word_features.json, where all
    possible values should be stored (unfortunately, stanfordnlp does not offer
    a comprehensive list of possible values for each model).

    If this transformer encounters an element not in that file, it will add one
    """

    def __init__(self):
        with open(WORD_FEATURES_JSON) as i_fh:
            self.features = json.load(i_fh)
            self.features.append(UNKNOWN_KEY)  # for unknown keys

    def fit(self, X, y=None, **fit_kwargs):
        return self

    def transform(self, X, y=None):
        result = []
        for document in X:
            doc = {}
            for feature in self.features:
                doc[feature] = 0
            for features_dict in _get_features_per_word(document):
                for k, v in features_dict.items():
                    key = f'{k}__{v}'
                    if key not in doc:
                        key = UNKNOWN_KEY
                        warnings.warn(
                            f'key {key} not found during fit, returning UNK')
                    doc[key] += 1
            result.append(doc)
        return result


class StanzaNlpToFieldTransformer(BaseEstimator, TransformerMixin):
    """
    Flattens a stanford document in the same order as the parsed text.

    Input (Document | Sentence):
        each document is expected to be a StanfordNLP document or sentence.
    Output: each document is returned as a single string, where each word is
        represented as the field that is to be extracted, separated by spaces.
        Sentences are separated by newlines.
    """

    def __init__(self, field=None):
        self.field = field

    def fit(self, x, y=None):
        return self

    def transform(self, x, y=None):
        result = []
        for document in x:
            document_result = []
            sentences: [Sentence] = []
            if type(document) == Sentence:
                sentences = [document]
            elif type(document) == Document:
                sentences = document.sentences
            for sentence in sentences:
                new_sentence = []
                for word in sentence.words:
                    label = _get_word_attribute(word, self.field, fallback='_')
                    if label is None:
                        raise ValueError(f'label is None for word: {word}')
                    new_sentence.append(label)
                document_result.append(' '.join(new_sentence))
            result.append('\n'.join(document_result))
        return result


class StanzaParserTransformer(BaseEstimator, TransformerMixin):
    """
    This transformer takes text and parses it using the stanza python
    package, which uses theano and various neural models to parse different
    features from natural text.
    """

    def __init__(self, language, silent=False, cpu=False):
        self.language = language
        self.silent = silent
        self.cpu = cpu
        if cpu:
            logger.info('using CPU for stanford parsing')
            os.environ['CUDA_VISIBLE_DEVICES'] = ''
        self.nlp = stanza.Pipeline(lang=self.language, use_gpu=not cpu)

    def fit(self, x, y=None):
        return self

    def transform(self, x, y=None):
        result = []
        if self.silent:
            pbar = x
        else:
            pbar = tqdm(x)
        for document in pbar:
            with open('/tmp/stanford_parsing', 'a') as log_fh:
                log_fh.write(document + '/n')
                result.append(self.nlp(document))
        return result


class StanzaToNltkTreesTransformer(BaseEstimator, TransformerMixin):
    """
    This transformer takes documents created by the stanza python package
    and transforms them into dependency trees in nltk format.

    Input: each document is either a stanza Document or Sentence.
    Output: each document is a list of NLTK Trees, one for each sentence.
        If the input document is a sentence, than the result is a list with
        a single entry.

    """

    def __init__(self, node_labels=None):
        """
        Initializes transformer.

        Args:
            node_labels (list of str): how each word is represented. Provide a
                list of ['lemma', 'upos', 'xpos', 'dependency'], e.g.:
                node_contents=['upos', 'dependency']
        """
        if type(node_labels) == str:
            node_labels = [node_labels]
        if not node_labels:
            node_labels = ['upos']
        self.node_labels = node_labels

    def fit(self, x, y=None):
        return self

    def transform(self, x, y=None):
        result = []
        for document in x:
            sentences = []
            if type(document) == Document:
                sentences = document.sentences
            elif type(document) == Sentence:
                sentences = [document]
            doc = []
            for sentence in sentences:
                dependency_tree = self.parse(sentence)
                labeled_tree = self.get_symbols(dependency_tree)
                doc.append(labeled_tree)
            if not doc:
                raise ValueError(
                    f'coud not parse anything from input {document}')
            result.append(doc)
        return result

    def get_symbols(self, tree):
        dependency = tree.label()
        if isinstance(dependency, str):
            return nltk.Tree(dependency, [self.get_symbols(c) for c in tree])
            # root case

        # the dependency consists of three parts:
        _, relationship, word_2 = dependency
        # the first word is already handled in other iterations of this method
        labels = []
        if 'dependency' in self.node_labels:
            labels.append(relationship)
        labels += [_get_word_attribute(word_2, label, fallback='_')
                   for label in self.node_labels
                   if label != 'dependency']
        label = '#'.join(labels)
        result = nltk.Tree(label, [self.get_symbols(c) for c in tree])
        return result

    def get_nodes(self, tree):
        leaves = []
        if isinstance(tree, nltk.Tree):
            if not isinstance(tree.label(), str):
                leaves.append(tree)
            for child in tree:
                leaves += self.get_nodes(child)
        return leaves

    def parse(self, sentence):
        roots = [d for d in sentence.dependencies if d[1] == 'root']
        unprocessed = [d for d in sentence.dependencies if d not in roots]
        # it is possible that multiple words have a root relationship
        tree = nltk.Tree('root', [nltk.Tree(x, []) for x in roots])

        while True:
            changed = False
            added_dependencies = []
            nodes = self.get_nodes(tree)
            for d in unprocessed:
                for leaf in nodes:
                    if hasattr(d[0], 'id'):
                        if leaf.label()[2].id == d[0].id:
                            leaf.append(nltk.Tree(d, []))
                            added_dependencies.append(d)
                            changed = True
                    elif hasattr(d[0], 'index'):
                        if leaf.label()[2].index == d[0].index:
                            leaf.append(nltk.Tree(d, []))
                            added_dependencies.append(d)
                            changed = True
            unprocessed = [x for x in unprocessed
                           if x not in added_dependencies]
            if len(unprocessed) == 0 or not changed:
                break

        return tree



from dstoolbox.transformers import Padder2d
from sklearn.base import TransformerMixin, BaseEstimator


class MaxLengthPadder(TransformerMixin, BaseEstimator):

    def __init__(self, pad_value, min_length=5):
        self.pad_value = pad_value
        self.min_length = min_length

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, documents):
        max_len = max([len(document) for document in documents]
                      + [self.min_length])
        print('max len is', max_len)
        padder = Padder2d(max_len=max_len, pad_value=self.pad_value)
        return padder.transform(documents)



import numpy as np
import pandas as pd
from simpletransformers.classification import ClassificationModel
from sklearn.base import BaseEstimator, TransformerMixin


class SimpletransformersBertModel(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        model_type='distilbert',
        pretrained_model='distilbert-base-uncased',
        output_dir='outputs/',
        cache_dir='cache/',
        fp16=False,
        fp16_opt_level='O1',
        max_seq_length=256,
        train_batch_size=8,
        eval_batch_size=8,
        gradient_accumulation_steps=1,
        num_train_epochs=1,
        weight_decay=0,
        learning_rate=4e-5,
        adam_epsilon=1e-8,
        warmup_ratio=0.06,
        warmup_steps=0,
        max_grad_norm=1.0,
        logging_steps=0,
        evaluate_during_training=False,
        save_steps=2000,
        eval_all_checkpoints=True,
        use_tensorboard=False,
        overwrite_output_dir=False,
        reprocess_input_data=True,
        use_gpu=False,
        use_cuda=False,
        n_gpu=1,
        silent=False,
        use_multiprocessing=True,
    ):

        self.model_type = model_type
        self.pretrained_model = pretrained_model
        self.output_dir = output_dir
        self.cache_dir = cache_dir
        self.fp16 = fp16
        self.fp16_opt_level = fp16_opt_level
        self.max_seq_length = max_seq_length
        self.train_batch_size = train_batch_size
        self.eval_batch_size = eval_batch_size
        self.gradient_accumulation_steps = gradient_accumulation_steps
        self.num_train_epochs = num_train_epochs
        self.weight_decay = weight_decay
        self.learning_rate = learning_rate
        self.adam_epsilon = adam_epsilon
        self.warmup_ratio = warmup_ratio
        self.warmup_steps = warmup_steps
        self.max_grad_norm = max_grad_norm
        self.logging_steps = logging_steps
        self.evaluate_during_training = evaluate_during_training
        self.save_steps = save_steps
        self.eval_all_checkpoints = eval_all_checkpoints
        self.use_tensorboard = use_tensorboard
        self.overwrite_output_dir = overwrite_output_dir
        self.reprocess_input_data = reprocess_input_data
        self.use_gpu = use_gpu
        self.use_cuda = use_cuda
        self.n_gpu = n_gpu
        self.silent = silent
        self.use_multiprocessing = use_multiprocessing

        self.model_args = {
            'model_type': self.model_type,
            'pretrained_model': self.pretrained_model,
            'output_dir': self.output_dir,
            'cache_dir': self.cache_dir,
            'fp16': self.fp16,
            'fp16_opt_level': self.fp16_opt_level,
            'max_seq_length': self.max_seq_length,
            'train_batch_size': self.train_batch_size,
            'eval_batch_size': self.eval_batch_size,
            'gradient_accumulation_steps': self.gradient_accumulation_steps,
            'num_train_epochs': self.num_train_epochs,
            'weight_decay': self.weight_decay,
            'learning_rate': self.learning_rate,
            'adam_epsilon': self.adam_epsilon,
            'warmup_ratio': self.warmup_ratio,
            'warmup_steps': self.warmup_steps,
            'max_grad_norm': self.max_grad_norm,
            'logging_steps': self.logging_steps,
            'evaluate_during_training': self.evaluate_during_training,
            'save_steps': self.save_steps,
            'eval_all_checkpoints': self.eval_all_checkpoints,
            'use_tensorboard': self.use_tensorboard,
            'overwrite_output_dir': self.overwrite_output_dir,
            'reprocess_input_data': self.reprocess_input_data,
            'use_gpu': self.use_gpu,
            'n_gpu': self.n_gpu,
            'silent': self.silent,
            'use_multiprocessing': self.use_multiprocessing,
        }

    def fit(self, x, y, *args, **kwargs):
        targets = set(y)
        n_classes = len(targets)
        for target in targets:
            if not (isinstance(target, int) or isinstance(target, np.int64)):
                raise TypeError(
                    'This model only works on integer targets, but you passed:'
                    f' {target} of type {type(target)}.'
                )
            if target >= n_classes:
                raise Exception(
                    f'target {target} is larger than n_classes ({n_classes}). '
                    'this is not supported by pytorch and will cause a very '
                    'hard to debug error.'
                )

        self.model = ClassificationModel(
            self.model_type,
            self.pretrained_model,
            num_labels=n_classes,
            args=self.model_args,
            use_cuda=self.use_cuda,
        )
        df = pd.DataFrame(zip(x, y))
        self.model.train_model(df)

    def predict(self, x, *args, **kwargs):
        result, raw = self.model.predict(x)
        return result

    @property
    def _estimator_type(self):
        return 'classifier'



from collections.abc import Iterable

import numpy as np
from gensim.models.doc2vec import Doc2Vec, TaggedDocument
from sklearn.base import BaseEstimator, TransformerMixin

# gensim is really loud.
import logging

logging.getLogger('gensim').setLevel(logging.WARNING)
logging.getLogger('gensim').propagate = False


def _sanity_check(documents):
    if len(documents) < 1:
        raise ValueError('empty document set provided')

    if type(documents[0]) == str or not isinstance(documents[1], Iterable):
        raise TypeError('this transformer only works on pre-split data')


class Doc2VecTransformer(TransformerMixin, BaseEstimator):
    """

    input: a list of list of strings
    output: a list of list of numbers

    """

    def __init__(
        self,
        learning_rate=0.02,
        epochs=20,
        vector_size=100,
        alpha=0.025,
        min_alpha=0.00025,
        min_count=2,
        dm=1,
        workers=1,
    ):
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.vector_size = vector_size
        self.alpha = alpha
        self.min_alpha = min_alpha
        self.min_count = min_count
        self.dm = dm
        self.workers = workers

        self.model = Doc2Vec(
            vector_size=vector_size,
            alpha=alpha,
            min_alpha=min_alpha,
            min_count=min_count,
            dm=dm,
            workers=workers,
        )

    def fit(self, documents, labels=None, **fit_params):
        documents = [
            [str(x) for x in document]
            for document in documents
        ]

        _sanity_check(documents)
        tagged_x = [
            TaggedDocument(words=row, tags=[])
            for _, row in enumerate(documents)
        ]
        self.model.build_vocab(tagged_x)
        self.model.train(tagged_x, total_examples=self.model.corpus_count,
                         epochs=self.epochs)
        return self

    def transform(self, documents):
        documents = [
            [str(x) for x in document]
            for document in documents
        ]
        _sanity_check(documents)
        vectors = [self.model.infer_vector(doc) for doc in documents]
        return np.array(vectors)



# This transformers are a re-implementation of the feature extraction method
# presented in Llorens(2016).

import random
from collections import defaultdict

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


class LifeVectorizer(BaseEstimator, TransformerMixin):
    def __init__(self, fragment_sizes=None, samples=200, sample_type="bow",
                 force=True):
        if fragment_sizes is None:
            fragment_sizes = [200, 500, 800, 1000, 1500, 2000, 3000, 4000]
        valid_sample_types = ["bow", "fragment", "both"]
        if sample_type not in valid_sample_types:
            raise ValueError(
                f"unknown sample type: {sample_type}. valid values: "
                f"{valid_sample_types}"
            )
        self.fragment_sizes = fragment_sizes
        self.samples = samples
        self.sample_type = sample_type
        self.force = force

    def fit(self, x, y=None):
        return self

    def sample(self, words, fragment_size, method):
        ret = []
        wordcount = len(words)
        if wordcount < fragment_size:
            if self.force:
                fragment_size = wordcount
            else:
                raise ValueError(
                    f"fragment size ({fragment_size}) is larger than document "
                    f"size ({wordcount}) for document starting with: \n\n"
                    f'{" ".join(words[:50])}\n\n'
                )
        for _ in range(self.samples):
            if method == "fragment":
                left = random.randint(0, wordcount - fragment_size)
                right = left + fragment_size
                ret.append(words[left:right])
            if method == "bow":
                ret.append(random.sample(words, fragment_size))
        return ret

    def get_features_for_sample(self, sample):
        counts = defaultdict(int)
        for word in sample:
            counts[word] += 1
        v0 = len(counts.keys())
        v1, v2, v3 = 0, 0, 0
        for occurrances in counts.values():
            if occurrances <= 1:
                v1 += 1
            elif occurrances <= 4:
                v2 += 1
            elif occurrances <= 10:
                v3 += 1

        return [v0, v1, v2, v3]

    def get_features(self, document, sample_size):
        if self.sample_type == "both":
            return np.concatenate(
                [
                    self._get_features(document, sample_size, "bow"),
                    self._get_features(document, sample_size, "fragment"),
                ]
            )
        else:
            return self._get_features(document, sample_size, self.sample_type)

    def _get_features(self, document, fragment_size, method):
        samples = self.sample(document, fragment_size, method)
        features = []
        for sample in samples:
            features.append(self.get_features_for_sample(sample))
        features = np.array(features)
        means = np.mean(features, axis=0)
        stds = np.std(features, axis=0)
        return np.concatenate(
            [means,
             np.divide(means, stds, out=np.zeros_like(means), where=stds != 0)]
        )

    def transform(self, x, y=None):
        ret = []
        for document in x:
            doc = [self.get_features(document, size) for size in
                   self.fragment_sizes]
            ret.append(np.concatenate(doc))

        # some classifiers like XGBoost require a numpy array if nested
        return np.array(ret)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tuhlbox']

package_data = \
{'': ['*']}

install_requires = \
['dstoolbox>=0.10.1,<0.11.0',
 'gensim>=3.8.3,<4.0.0',
 'simpletransformers>=0.51.1,<0.52.0',
 'sklearn>=0.0,<0.1',
 'stanza>=1.1.1,<2.0.0',
 'torch>=1.7.0,<2.0.0',
 'treegrams>=0.1.0,<0.2.0']

setup_kwargs = {
    'name': 'tuhlbox',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Benjamin Murauer',
    'author_email': 'b.murauer@posteo.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4',
}


setup(**setup_kwargs)
